# vim: fileencoding=UTF-8 filetype=python ff=unix expandtab sw=4 sts=4 tw=120
# maintainer: Christer Sjöholm -- goobook AT furuvik DOT net
# authors: Marcus Nitzschke -- marcus.nitzschke AT gmx DOT com
#
# Copyright (C) 2009  Carlos José Barroso
# Copyright (C) 2010  Christer Sjöholm
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''\
The idea is make an interface to google contacts that mimics the behaviour of
abook for mutt. It's developed in python and uses the fine
google data api (gdata).
'''

import codecs
import email.parser
import email.header
import gdata.service
import httplib2
import itertools
import logging
import os
import pickle
import re
import sys
import time
import xml.etree.ElementTree as ET

from .storage import Storage, storageify, unstorageify

log = logging.getLogger(__name__)

CACHE_FORMAT_VERSION = '4.0'
G_MAX_SRESULTS = 9999  # Maximum number of entries to ask google for.
GDATA_VERSION = '3'
ATOM_NS = '{http://www.w3.org/2005/Atom}'
G_NS = '{http://schemas.google.com/g/2005}'
GC_NS = '{http://schemas.google.com/contact/2008}'


class GooBook(object):
    '''This class can't be used as a library as it looks now, it uses sys.stdin
       print and sys.exit().'''
    def __init__(self, config):
        self.__config = config
        self.cache = Cache(config)
        self.cache.load()

    def query(self, query):
        """Do the query, and print it out in

        """
        # query contacts
        matching_contacts = sorted(self.__query_contacts(query), key=lambda c: c.title)
        # query groups
        matching_groups = sorted(self.__query_groups(query), key=lambda g: g.title)
        # mutt's query_command expects the first line to be a message,
        # which it discards.
        print "\n",
        for contact in matching_contacts:
            if contact.emails:
                emailaddrs = sorted(contact.emails)
                groups = set(self.cache.get_group(gid).title for gid in contact.groups)
                groups = groups.difference(('System Group: My Contacts',))
                groups_str = ', '.join(('"' + g + '"' for g in groups))
                for (emailaddr, kind) in emailaddrs:
                    title = contact.title or contact.nickname or emailaddr
                    extra_str = kind
                    if groups_str:
                        extra_str = extra_str + ' groups: ' + groups_str
                    print (u'\t'.join((emailaddr, title, extra_str))).encode(self.__config.encoding, errors='replace')
        for group in matching_groups:
            emails = ['%s <%s>' % (c.title, c.emails[0][0]) for c in group.contacts if c.emails]
            emails = ', '.join(emails)
            if not emails:
                continue
            print (u'%s\t%s (group)' % (emails, group.title)).encode(self.__config.encoding, errors='replace')

    def query_details(self, query):
        """
        Method for querying the contacts and printing
        a detailed view.
        """

        out = codecs.getwriter(self.__config.encoding)(sys.stdout, errors='replace')

        # query contacts
        matching_contacts = sorted(self.__query_contacts(query), key=lambda c: c.title)
        # query groups
        matching_groups = sorted(self.__query_groups(query), key=lambda g: g.title)
        for group in matching_groups:
            matching_contacts += group.contacts

        for contact in matching_contacts:
            print >> out, "-------------------------"
            print >> out, contact.title
            if contact.birthday:
                print >> out, "Birthday: ", contact.birthday
            if contact.phonenumbers:
                print >> out, "Phone:"
                for (number, kind) in contact.phonenumbers:
                    print >> out, "\t", number, " (" + kind + ")"
            if contact.emails:
                print >> out, "EMail:"
                emailaddrs = sorted(contact.emails)
                for (emailaddr, kind) in emailaddrs:
                    print >> out, "\t", emailaddr, " (" + kind + ")"
            if contact.im:
                print >> out, "IM:"
                for (nick, protocol) in contact.im:
                    print >> out, "\t", nick, " (", protocol, ")"
            if contact.addresses:
                print >> out, "Address:"
                for (address, kind) in contact.addresses:
                    lines = address.splitlines()
                    lines[0] = '%s ( %s )' % (lines[0], kind)
                    print >> out, "\t" + '\n\t'.join(lines)
            if contact.groups:
                print >> out, "Groups:"
                groups = set(self.cache.get_group(gid).title for gid in contact.groups)
                groups = groups.difference(('System Group: My Contacts',))
                groups_str = '\n\t'.join(groups)
                print >> out, "\t" + groups_str

    def __query_contacts(self, query):
        match = re.compile(query, re.I).search  # create a match function
        for contact in self.cache.contacts:
            if self.__config.filter_groupless_contacts and not contact.groups:
                continue  # Skip contacts without groups
            if any(itertools.imap(match,
                                  [contact.title, contact.nickname] + [unicode(number) for
                                                                       (number, kind) in contact.phonenumbers])):
                yield contact
            else:
                matching_addrs = [(email, kind) for (email, kind) in contact.emails if match(email)]
                if matching_addrs:
                    contact.emails = matching_addrs  # only show matching
                    yield contact

    def __query_groups(self, query):
        match = re.compile(query, re.I).search  # create a match function
        for group in self.cache.groups:
            # Collect all values to match against
            all_values = (group.title,)
            if any(itertools.imap(match, all_values)):
                group.contacts = list(self.__get_group_contacts(group.id))
                yield group

    def __get_group_contacts(self, group_id):
        for contact in self.cache.contacts:
            if group_id in contact.groups:
                yield contact

    def add_mail_contact(self, name, mailaddr):
        entry = ET.Element(ATOM_NS + 'entry')
        ET.SubElement(entry, ATOM_NS + 'category', scheme='http://schemas.google.com/g/2005#kind',
                      term='http://schemas.google.com/contact/2008#contact')
        fullname_e = ET.Element(G_NS + 'fullName')
        fullname_e.text = name
        ET.SubElement(entry, G_NS + 'name').append(fullname_e)
        ET.SubElement(entry, G_NS + 'email', rel='http://schemas.google.com/g/2005#other', primary='true',
                      address=mailaddr)

        group_id = self.cache.get_group_by_title('System Group: My Contacts').id
        ET.SubElement(entry, GC_NS + 'groupMembershipInfo', deleted='false',
                      href=group_id)

        if self.__config.default_group:
            group_id2 = self.cache.get_group_by_title(self.__config.default_group).id
            ET.SubElement(entry, GC_NS + 'groupMembershipInfo', deleted='false',
                          href=group_id2)

        gcont = GoogleContacts(self.__config)
        log.debug('Going to create contact name: %s email: %s' % (name, mailaddr))
        gcont.create_contact(entry)
        log.info('Created contact name: %s email: %s' % (name, mailaddr))

    def add_email_from(self, lines):
        """Add an address from From: field of a mail.
        This assumes a single mail file is supplied through.

        Args:
          lines: A generator of lines, usually a open file.

        """

        parser = email.parser.HeaderParser()
        headers = parser.parse(lines)
        if 'From' not in headers:
            print "Not a valid mail file!"
            sys.exit(2)

        (name, mailaddr) = email.utils.parseaddr(headers['From'])
        if not name:
            name = mailaddr
        else:
            # This decodes headers like "=?iso-8859-1?q?p=F6stal?="
            values = email.header.decode_header(name)
            if len(values) == 0:
                # Can't this be possible?
                name = mailaddr
            else:
                # There should be only one element anyway
                (name, encoding) = values[0]

                if encoding is not None:
                    name = name.decode(encoding)

        self.add_mail_contact(name, mailaddr)


class Cache(object):
    def __init__(self, config):
        self.__config = config
        self.contacts = None  # list of Storage
        self.groups = None  # list of Storage

    def load(self, force_update=False):
        """Load the cached addressbook feed, or fetch it (again) if it is
        old or missing or invalid or anyting

        Args:
          force_update: force update of cache

        """
        cache = {}

        # if cache newer than cache_expiry_hours
        if not force_update and (os.path.exists(self.__config.cache_filename) and
                                 ((time.time() - os.path.getmtime(self.__config.cache_filename)) <
                                  (int(self.__config.cache_expiry_hours) * 60 * 60))):
            try:
                log.debug('Loading cache: ' + self.__config.cache_filename)
                cache = pickle.load(open(self.__config.cache_filename, 'rb'))
                if cache.get('goobook_cache') != CACHE_FORMAT_VERSION:
                    log.info('Detected old cache format')
                    cache = None  # Old cache format
            except StandardError, err:
                log.info('Failed to read the cache file: %s', err)
                raise
        if cache:
            self.contacts = storageify(cache.get('contacts'))
            self.groups = storageify(cache.get('groups'))
        else:
            self.update()
        if not self.contacts:
            raise Exception('Failed to find any contacts')  # TODO

    def update(self):
        log.info('Retrieving contact data from Google.')
        gc = GoogleContacts(self.__config)
        self.contacts = list(self._parse_contacts(gc.fetch_contacts()))
        self.groups = list(self._parse_groups(gc.fetch_contact_groups()))
        self.save()

    def save(self):
        """Pickle the addressbook and a timestamp

        """
        if self.contacts:  # never write a empty addressbook
            cache = {'contacts': unstorageify(self.contacts),
                     'groups': unstorageify(self.groups),
                     'goobook_cache': CACHE_FORMAT_VERSION}
            pickle.dump(cache, open(self.__config.cache_filename, 'wb'))

    def get_group(self, id_):
        for group in self.groups:
            if group.id == id_:
                return group
        raise KeyError('Group: ' + id_)

    def get_group_by_title(self, title):
        for group in self.groups:
            if group.title == title:
                return group
        raise KeyError('Group: ' + title)

    @staticmethod
    def _parse_contact(entry):
        '''Extracts interesting contact info from cache.'''
        contact = Storage()
        # ID
        contact.id = entry.findtext(ATOM_NS + 'id')
        # title
        contact.title = entry.findtext(ATOM_NS + 'title')
        # nickname
        contact.nickname = entry.findtext(GC_NS + 'nickname', default='')
        # emails
        contact.emails = []
        for ent in entry.findall(G_NS + 'email'):
            label = ent.get('label') or ent.get('rel').split('#')[-1]
            contact.emails.append((ent.get('address'), label))
        # groups
        contact.groups = [e.get('href') for e in entry.findall(GC_NS + 'groupMembershipInfo') if
                          e.get('deleted') == 'false']
        # phone
        contact.phonenumbers = []
        for ent in entry.findall(G_NS + 'phoneNumber'):
            label = ent.get('label') or ent.get('rel').split('#')[-1]
            contact.phonenumbers.append((ent.text, label))
        # birthday
        contact.birthday = entry.find(GC_NS + 'birthday').get('when') if entry.findall(GC_NS + 'birthday') else None
        # address
        contact.addresses = []
        for address in entry.findall(G_NS + 'structuredPostalAddress'):
            label = address.get('label') or address.get('rel').split('#')[-1]
            contact.addresses.append((address.findtext(G_NS + 'formattedAddress'), label))
        # IM
        contact.im = []
        for ent in entry.findall(G_NS + 'im'):
            protocol = ent.get('protocol')
            # Default protocol is GOOGLE_TALK
            protocol = ent.get('protocol').split('#')[-1] if protocol else "GOOGLE_TALK"
            contact.im.append((ent.get('address'), protocol))

        log.debug('Parsed contact %s', contact)
        return contact

    @staticmethod
    def _parse_group(entry):
        '''Extracts interesting group info from cache.'''
        group = Storage()
        group.id = entry.findtext(ATOM_NS + 'id')
        group.title = entry.findtext(ATOM_NS + 'title')
        log.debug('Parsed group %s', group)
        return group

    def _parse_contacts(self, raw_contacts):
        for entry in raw_contacts.findall(ATOM_NS + 'entry'):
            yield self._parse_contact(entry)

    def _parse_groups(self, raw_groups):
        for entry in raw_groups.findall(ATOM_NS + 'entry'):
            yield self._parse_group(entry)


class GoogleContacts(object):

    def __init__(self, config):
        self.__http_client = self.__get_client(config.creds)
        self.__additional_headers = {
            'GData-Version': GDATA_VERSION,
            'Content-Type': 'application/atom+xml'
        }

    def __get_client(self, credentials):
        '''Login to Google and return a ContactsClient object.

        '''
        if not credentials or credentials.invalid:
            sys.exit(u'No or invalid credentials, run "goobook authenticate"')  # TODO raise exception instead
        http_auth = credentials.authorize(httplib2.Http())
        return http_auth

    def _get(self, query):
        resp_headers, content = self.__http_client.request(str(query),
                                                           'GET',
                                                           headers=self.__additional_headers,
                                                           connection_type=httplib2.HTTPSConnectionWithTimeout)
        log.debug('GET returned: %s', resp_headers)
        if resp_headers['status'] != '200':
            raise StandardError('Failed headers: {} content: {}'.format(resp_headers, content))
        res = ET.fromstring(content)
        return res

    def _post(self, data, query):
        '''data is a ElementTree'''
        data = ET.tostring(data)
        log.debug('POSTing to: %s\n%s', query, data)
        resp_headers, content = self.__http_client.request(str(query),
                                                           'POST',
                                                           data,
                                                           headers=self.__additional_headers,
                                                           connection_type=httplib2.HTTPSConnectionWithTimeout)
        if resp_headers['status'] != '201':
            raise StandardError('Failed headers: {} content: {}'.format(resp_headers, content))
        log.debug('POST returned: %s', resp_headers)
        # res = self.__client.Post(data, str(query), converter=str)

    def fetch_contacts(self):
        query = gdata.service.Query('http://www.google.com/m8/feeds/contacts/default/full')
        query.max_results = G_MAX_SRESULTS
        res = self._get(query)
        return res

    def fetch_contact_groups(self):
        query = gdata.service.Query('http://www.google.com/m8/feeds/groups/default/full')
        query.max_results = G_MAX_SRESULTS
        res = self._get(query)
        return res

    def create_contact(self, entry):
        query = gdata.service.Query('http://www.google.com/m8/feeds/contacts/default/full')
        self._post(entry, query)



# -*- coding: UTF-8 -*-
# vim: fileencoding=UTF-8 filetype=python ff=unix et ts=4 sw=4 sts=4 tw=120
# author: Christer Sjöholm -- hcs AT furuvik DOT net

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import collections
import json


class Storage(object):
    """
    A Storage object wraps a dictionary.
    In addition to `obj['foo']`, `obj.foo` can also be used for accessing
    values.

    Wraps the dictionary instead of extending it so that the number of names
    that can conflict with keys in the dict is kept minimal.

        >>> o = Storage(a=1)
        >>> o.a
        1
        >>> o['a']
        1
        >>> 'a' in o
        True
        >>> o.a = 2
        >>> o['a']
        2
        >>> del o.a
        >>> o.a
        Traceback (most recent call last):
            ...
        AttributeError: 'a'

        Based on Storage in web.py (public domain)
    """
    __internal_vars = ('_dict', '_normalize', '_denormalize')

    def __init__(self, dict_=None, default_factory=None, normalize=None,
                 denormalize=None, case_insensitive=False, **kwargs):
        '''
        dict_: Dict to use as backend for Storage object, normalize will not
               be called for existing items.
        case_insensitive: All given keys will be converted to lower case before
                          trying to set/access the dict. If a populated dict_
                          is given it must already have all keys in lower case.
        default_factory: if dict_ is None and default_factory is set then a
                         defaultdict will be used with the specified
                         default_factory.
        normalize: function that normalizes keys, case_insensitive for example
                   is implemented as a normalizer that lower cases each key.
        denormalize: function that is called on each key before it is returned
                     to the caller(iteritems and __iter__).

        '''
        if dict_ is not None:
            self._dict = dict_
        elif default_factory:
            self._dict = collections.defaultdict(default_factory, **kwargs)
        else:
            self._dict = dict(**kwargs)
        if normalize and case_insensitive:
            self._normalize = lambda key: normalize(key.lower())
        elif case_insensitive:
            self._normalize = lambda key: key.lower()
        elif normalize:
            self._normalize = normalize
        else:
            self._normalize = lambda key: key  # Do nothing
        if denormalize:
            self._denormalize = denormalize
        else:
            self._denormalize = lambda key: key  # Do nothing
        for key, value in kwargs.items():
            self[key] = value

    def get_dict(self):
        ''' Get the contained dict.
            all keys will be in their normalized form.
        '''
        return self._dict

    def iteritems(self):
        ''' Iterate over all (key, value) pairs.
        All keys will be denormalized.
        '''
        for key, value in self._dict.items():
            yield self._denormalize(key), value

    def __getattr__(self, key):
        try:
            # prevent recursion (triggered by pickle.load()
            if key in Storage.__internal_vars:
                raise AttributeError('No such attribute: ' + repr(key))
            key = self._normalize(key)
            return self[key]
        except KeyError as err:
            raise AttributeError(err)

    def __setattr__(self, key, value):
        # prevent recursion (triggered by pickle.load()
        if key in Storage.__internal_vars:
            object.__setattr__(self, key, value)
        else:
            key = self._normalize(key)
            self[key] = value

    def __delattr__(self, key):
        try:
            key = self._normalize(key)
            del self[key]
        except KeyError as err:
            raise AttributeError(err)

    # For container methods pass-through to the underlying dict.
    def __getitem__(self, key):
        key = self._normalize(key)
        return self._dict[key]

    def __setitem__(self, key, value):
        key = self._normalize(key)
        self._dict[key] = value

    def __delitem__(self, key):
        key = self._normalize(key)
        del self._dict[key]

    def __contains__(self, key):
        key = self._normalize(key)
        return key in self._dict

    def __iter__(self):
        for key in self._dict:
            yield self._denormalize(key)

    def __len__(self):
        return self._dict.__len__()

    def __eq__(self, other):
        return isinstance(other, Storage) and self._dict == other._dict

    __hash__ = None

    def __repr__(self):
        items = []
        if isinstance(self._dict, collections.defaultdict):
            items.append('default_factory={0}'.format(self._dict.default_factory))
        items.extend('{0}={1}'.format(self._denormalize(i[0]), repr(i[1]))
                     for i in sorted(self._dict.items()))
        return '{0}({1})'.format(self.__class__.__name__, ', '.join(items))


def json_loads_storage(str_):
    '''
        >>> json_loads_storage('[{"a":1}]')
        [Storage(a=1)]
    '''
    return json.loads(str_, object_hook=Storage)


def json_load_storage(fp_):
    return json.load(fp_, object_hook=Storage)


def json_dumps_storage(job):
    '''
        >>> json_dumps_storage(Storage(a=2))
        '{\\n  "a": 2\\n}'
        >>> json_dumps_storage([Storage(a=2)])
        '[\\n  {\\n    "a": 2\\n  }\\n]'
        >>> json_dumps_storage({'a':2})
        '{\\n  "a": 2\\n}'
    '''
    return json.dumps(job, default=_storage_to_dict, indent=2)


def json_dump_storage(obj, fp_):
    return json.dump(obj, fp_, default=_storage_to_dict, indent=2)


def _storage_to_dict(obj):
    '''
        >>> _storage_to_dict(Storage(a=1))
        {'a': 1}
        >>> _storage_to_dict('')
        Traceback (most recent call last):
        ...
        TypeError: not a Storage object

    '''
    if isinstance(obj, Storage):
        return obj.get_dict()
    else:
        raise TypeError('not a Storage object')


def storageify(obj, storageFactory=Storage):
    '''
    takes a json style datastructure and converts all dicts to Storage.
    '''
    if isinstance(obj, dict):
        res = storageFactory()
        for key, value in obj.items():
            res[key] = storageify(value, storageFactory=storageFactory)
    elif isinstance(obj, list):
        res = []
        for element in obj:
            res.append(storageify(element, storageFactory=storageFactory))
    else:
        res = obj
    return res


def unstorageify(obj):
    '''
    make a deep copy of Storage, dict, and lists
    and convert all Storage to dict
    Good when you want to convert to json or yaml
    '''
    if isinstance(obj, Storage):
        res = {}
        for key, value in obj.get_dict().items():
            res[key] = unstorageify(value)
    elif isinstance(obj, dict):
        res = {}
        for key, value in obj.items():
            res[key] = unstorageify(value)
    elif isinstance(obj, list):
        res = []
        for element in obj:
            res.append(unstorageify(element))
    else:
        res = obj
    return res



#!/usr/bin/env python
# vim: fileencoding=UTF-8 filetype=python ff=unix expandtab sw=4 sts=4 tw=120
# author: Christer Sjöholm -- goobook AT furuvik DOT net

import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()
NEWS = open(os.path.join(here, 'CHANGES.txt')).read()


class UltraMagicString(object):
    ''' Stolen from http://stackoverflow.com/questions/1162338/whats-the-right-way-to-use-unicode-metadata-in-setup-py

    Catch-22:
    - if I return Unicode, python setup.py --long-description as well
      as python setup.py upload fail with a UnicodeEncodeError
    - if I return UTF-8 string, python setup.py sdist register
      fails with an UnicodeDecodeError
    '''

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value

    def __unicode__(self):
        return self.value.decode('UTF-8')

    def __add__(self, other):
        return UltraMagicString(self.value + str(other))

    def split(self, *args, **kw):
        return self.value.split(*args, **kw)

setup(name='goobook',
      version='1.8',
      description='Search your google contacts from the command-line or mutt.',
      long_description=UltraMagicString(README + '\n\n' + NEWS),
      maintainer=UltraMagicString('Christer Sjöholm'),
      maintainer_email='goobook@furuvik.net',
      url='http://gitlab.com/goobook/goobook',
      download_url='http://pypi.python.org/pypi/goobook',
      classifiers=[f.strip() for f in """
        Development Status :: 5 - Production/Stable
        Environment :: Console
        Operating System :: OS Independent
        Programming Language :: Python
        Programming Language :: Python :: 2.7
        Intended Audience :: End Users/Desktop
        License :: OSI Approved :: GNU General Public License (GPL)
        Topic :: Communications :: Email :: Address Book
        """.splitlines() if f.strip()],
      license='GPLv3',
      install_requires=[
          'oauth2client>=1.4.11',
          'setuptools>=0.7',
          'gdata>=2.0.18',
          'simplejson>=2.1.0',
          'httplib2>=0.9.1'],
      extras_require={
          'keychain': ['keyring>=0.2']
      },
      include_package_data=True,
      zip_safe=False,
      packages=find_packages(),
      entry_points={'console_scripts': ['goobook = goobook.application:main']}
      )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# vim: fileencoding=UTF-8 filetype=python ff=unix et ts=4 sw=4 sts=4 tw=120
# author: Christer Sjöholm -- hcs AT furuvik DOT net

from __future__ import absolute_import
from pkg_resources import resource_filename
import argparse
import goobook.config
import logging
import oauth2client.client
import oauth2client.file
import oauth2client.tools
import os
import pkg_resources
import sys
import xml.etree.ElementTree as ElementTree

from goobook.goobook import GooBook, Cache, GoogleContacts

log = logging.getLogger(__name__)

CONFIG_FILE = '~/.goobookrc'
SCOPES = 'https://www.google.com/m8/feeds'  # read/write access to Contacts and Contact Groups


def main():
    parser = argparse.ArgumentParser(description='Search you Google contacts from mutt or the command-line.')
    parser.add_argument('-c', '--config', help='Specify alternative configuration file.', metavar="FILE")
    parser.add_argument('-v', '--verbose', dest="logging_level", action='store_const',
                        const=logging.INFO, help='Be verbose about what is going on (stderr).')
    parser.add_argument('-V', '--version',
                        action='version',
                        version='%%(prog)s %s' % pkg_resources.get_distribution("goobook").version,
                        help="Print version and exit")
    parser.add_argument('-d', '--debug', dest="logging_level", action='store_const',
                        const=logging.DEBUG, help='Output debug info (stderr).')
    parser.set_defaults(config=CONFIG_FILE, logging_level=logging.ERROR)

    subparsers = parser.add_subparsers()

    parser_add = subparsers.add_parser('add',
                                       description='Create new contact, if name and email is not given the'
                                       ' sender of a mail read from stdin will be used.')
    parser_add.add_argument('name', nargs='?', metavar='NAME',
                            help='Name to use.')
    parser_add.add_argument('email', nargs='?', metavar='EMAIL',
                            help='E-mail to use.')
    parser_add.set_defaults(func=do_add)

    parser_config_template = subparsers.add_parser('config-template',
                                                   description='Prints a template for .goobookrc to stdout')
    parser_config_template.set_defaults(func=do_config_template)

    parser_dump_contacts = subparsers.add_parser('dump_contacts',
                                                 description='Dump contacts as XML.')
    parser_dump_contacts.set_defaults(func=do_dump_contacts)

    parser_dump_groups = subparsers.add_parser('dump_groups',
                                               description='Dump groups as XML.')
    parser_dump_groups.set_defaults(func=do_dump_groups)

    parser_query = subparsers.add_parser('query',
                                         description='Search contacts using query (regex).')
    parser_query.add_argument('query', help='regex to search for.', metavar='QUERY')
    parser_query.set_defaults(func=do_query)

    parser_query_details = subparsers.add_parser(
        'dquery',
        description='Search contacts using query (regex) and print out all info.')
    parser_query_details.add_argument('query', help='regex to search for.')
    parser_query_details.set_defaults(func=do_query_details)

    parser_reload = subparsers.add_parser('reload',
                                          description='Force reload of the cache.')
    parser_reload.set_defaults(func=do_reload)

    parser_reload = subparsers.add_parser('authenticate',
                                          description='Google OAuth authentication.',
                                          formatter_class=argparse.RawDescriptionHelpFormatter,
                                          parents=[oauth2client.tools.argparser])
    parser_reload.set_defaults(func=do_authenticate)

    args = [arg.decode(goobook.config.ENCODING) for arg in sys.argv[1:]]
    args = parser.parse_args(args)

    logging.basicConfig(level=args.logging_level)

    try:
        if args.func == do_config_template:
            config = None
        else:
            config = goobook.config.read_config(args.config)
        args.func(config, args)
    except goobook.config.ConfigError as err:
        sys.exit(u'Configuration error: ' + unicode(err))

##############################################################################
# sub commands


def do_add(config, args):
    goobk = GooBook(config)
    if args.name and args.email:
        goobk.add_mail_contact(args.name, args.email)
    else:
        goobk.add_email_from(sys.stdin)


def do_config_template(config, args):
    print goobook.config.TEMPLATE


def do_dump_contacts(config, args):
    goco = GoogleContacts(config)
    print ElementTree.tostring(goco.fetch_contacts(), 'UTF-8')


def do_dump_groups(config, args):
    goco = GoogleContacts(config)
    print ElementTree.tostring(goco.fetch_contact_groups(), 'UTF-8')


def do_query(config, args):
    goobk = GooBook(config)
    goobk.query(args.query)


def do_query_details(config, args):
    goobk = GooBook(config)
    goobk.query_details(args.query)


def do_reload(config, args):
    cache = Cache(config)
    cache.load(force_update=True)


def do_authenticate(config, args):
    store = config.store
    creds = config.creds

    if not creds or creds.invalid:
        client_secret_filename = config.client_secret_filename
        if not os.path.exists(client_secret_filename):
            client_secret_filename = resource_filename(__name__, 'client_secret.json')
        flow = oauth2client.client.flow_from_clientsecrets(client_secret_filename, SCOPES)
        creds = oauth2client.tools.run_flow(flow, store, args)
    else:
        print 'You are already authenticated.'


if __name__ == '__main__':
    main()
